package com.jh.boot.security.service;

import java.util.Collection;
import java.util.Date;
import java.util.Locale;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import com.jh.boot.security.AppUserAuthentication;
import com.jh.boot.security.AuthService;
import com.jh.boot.security.AuthServiceListener;
import com.jh.boot.security.PasswordHash;
import com.jh.boot.security.PasswordUtils;
import com.jh.boot.security.model.AppRole;
import com.jh.boot.security.model.AppUser;
import com.jh.boot.security.model.ResetPasswordToken;
import com.jh.boot.security.repository.AppUserRepository;



public class AppUserAuthService implements AuthService {

	private static final Logger LOG = LoggerFactory.getLogger(AppUserAuthService.class);
	
	private AppUserRepository appUserRepository;
	
	private Collection<AuthServiceListener> authListeners;
	

	@Override
	@Transactional
	public Authentication authenticate(String login, String password) throws BadCredentialsException {
		AppUser user = appUserRepository.fetchByLoginDetached(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		if (!PasswordUtils.checkPassword(password, user.getPassword(), user.getPasswordSalt())) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}
		
		//sanitize object
		user.setDeleted(null);
		user.setPassword(null);
		user.setPasswordSalt(null);
		user.setVersion(null);
		
		//ADD implicit user
		user.getRoles().add(new AppRole("USER"));
		
		//initialize roles
		user.initRolesMap();
		
		Set<SimpleGrantedAuthority> authorities = user.getRoles().stream().map(r -> new SimpleGrantedAuthority("ROLE_" + r.getName())).collect(Collectors.toSet());
		AppUserAuthentication auth = new AppUserAuthentication(user, authorities);
		
		return auth;
	}

	@Override
	@Transactional
	public void register(String login, String password, Locale locale) throws AuthenticationException {
		AppUser user = appUserRepository.findByLogin(login);
		if (user != null) {
			throw new BadCredentialsException("AUTH.USER_ALREADY_EXISTS");
		}
		if (!StringUtils.hasText(login) || !StringUtils.hasText(password)) {
			throw new BadCredentialsException("AUTH.BAD_USERNAME_OR_PASSWORD");
		}
		if (!PasswordUtils.validatePassword(password)) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}
		AppUser appUser = appUserRepository.registerUser(login, password);
		
		if (authListeners != null) {
			authListeners.forEach(al -> al.registerUser(appUser));
		}
	}

	@Override
	@Transactional
	public String generateResetToken(String login) {
		AppUser user = appUserRepository.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		ResetPasswordToken token = new ResetPasswordToken(login, new Date(), UUID.randomUUID().toString());
		appUserRepository.saveResetPasswordToken(token);
		
		if (authListeners != null) {
			authListeners.forEach(al -> al.generateResetToken(user, token));
		}
		
		return token.getToken();
	}

	
	@Override
	@Transactional
	public void resetPassword(String login, String token, String newPassword) throws AuthenticationException {
		ResetPasswordToken rpt = appUserRepository.findResetPasswordToken(login, token);
		if (rpt == null) {
			throw new BadCredentialsException("AUTH.INVALID_TOKEN");
		}
	
		AppUser user = appUserRepository.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		PasswordHash hash = PasswordUtils.hashPassword(newPassword);
		user.setPassword(hash.getHash());
		user.setPasswordSalt(hash.getSalt());
		
		user.setPasswordChanged(new Date());
		user.setPasswordExpired(false);
	}
	
	
	@Override
	@Transactional
	public void delete(AppUser user) throws AuthenticationException {
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		user = appUserRepository.findByLogin(user.getEmail());
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		user.setDeleted(true);
		user.setAccountEnabled(false);
	}
	
	
	@Autowired(required = false)
	public void setAppUserRepository(AppUserRepository appUserRepository) {
		this.appUserRepository = appUserRepository;
	}

	@Autowired(required = false)
	public void setAuthListeners(Collection<AuthServiceListener> authListeners) {
		this.authListeners = authListeners;
	}

	@Override
	public void changePassword(String login, String currentPassword, String newPassword) throws AuthenticationException {
		AppUser user = appUserRepository.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		if (!PasswordUtils.checkPassword(currentPassword, user.getPassword(), user.getPasswordSalt())) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}
		
		PasswordHash hash = PasswordUtils.hashPassword(newPassword);
		user.setPassword(hash.getHash());
		user.setPasswordSalt(hash.getSalt());
		
	}
	
}
